#version 430

#define PI 3.1415926538
#define EPSILON .00001



//###################################################//
//### Inputs/uniforms for our gbuffer pixelshader ###//
//###################################################//
layout(std140, binding = 0) uniform perFrame
{
	mat4x4	vp;
	mat4x4  invVP;

	mat4x4	view;
	mat4x4	invView;
	mat4x4	invTransView;

	mat4x4	proj;
	mat4x4	invProj;

	vec3	eyePos;
	float	gt;
	float	dt;
	uint	screenW;
	uint	screenH;

};

layout(binding = 0) uniform sampler2D texAlbedoMetal;
layout(binding = 1) uniform sampler2D texNormalDepth;
layout(binding = 2) uniform sampler2D texEmissiveRoughness;
layout(binding = 4) uniform sampler2D texBackbuffer;
layout(binding = 5) uniform samplerCube texSkybox;

uniform float MAX_RAYLENGTH = 15.f; //in meters
uniform float NEAR_Z = 0.01f;
uniform float FAR_Z = 1000.f;
uniform float Z_THICKNESS = .015f; //in meters
uniform uint MAX_STEPS = 25;
uniform bool PASSTHROUGH = false;
uniform bool isHyperbolicZBuffer = true; //true: Zbuffer stores perspective depth; false: stores linear-z values



layout(location = 0) in vec2 inTex;

layout(location = 0) out vec4 outColor;





float distanceSquared(vec3 A, vec3 B) {
    A -= B;
    return dot(A, A);
}

float distanceSquared(vec2 A, vec2 B) {
    A -= B;
    return dot(A, A);
}


float linearizeZ(float z, float nearZ, float farZ)
{
	//linearize zbuffer after: 
	// https://stackoverflow.com/questions/42509883/how-to-correctly-linearize-depth-in-opengl-es-in-ios/42515399#42515399
	return( (z - nearZ) / (farZ - nearZ) );
}


void swap(inout float a, inout float b) {
     float temp = a;
     a = b;
     b = temp;
}


float limitRaylengthToZPlanesAndScreen(float maxRayLength, vec3 start, vec3 ray, float nearZ, float farZ, vec2 csScreen)
{
	float raylength = maxRayLength;
	
	const float csEndZ = start.z + ray.z * maxRayLength;
	const float raylengthBack	= abs(farZ - start.z) / abs(ray.z); //farZ = startZ + rayZ * raylength
	const float raylengthFront	= abs(start.z - nearZ) / abs(ray.z); //nearZ = startZ + rayZ * raylength //nearZ IS lesser than startZ

	const float raylengthRight	= abs(+csScreen.x - start.x) / abs(ray.x);
	const float raylengthLeft	= abs(-csScreen.x - start.x) / abs(ray.x);

	const float raylengthTop	= abs(+csScreen.y - start.y) / abs(ray.y);
	const float raylengthBot	= abs(-csScreen.y - start.y) / abs(ray.y);

	raylength = min(raylength, raylengthFront);
	raylength = min(raylength, raylengthBack);
	raylength = min(raylength, raylengthRight);
	raylength = min(raylength, raylengthLeft);
	raylength = min(raylength, raylengthTop);
	raylength = min(raylength, raylengthBot);

	return( raylength );
}


bool raymarch(vec3 csStart, vec3 csRay, float stepScale, float jitter, out vec2 hitPixel)
{
	vec2 csScreen = vec2(screenW, screenH);
	csScreen = (invProj * vec4(csScreen, 0, 0)).xy;

	//calculate raylength clipped to near- & far-plane
	const float raylength = limitRaylengthToZPlanesAndScreen(MAX_RAYLENGTH, csStart, csRay, NEAR_Z, FAR_Z, csScreen);
	vec3 csEnd = csStart + csRay * raylength;
	if( distanceSquared(csStart, csEnd) < EPSILON ) csEnd += 2.0f * EPSILON;


	//----------------------//
	//---Start-/Endpoints---//
	//----------------------//
	vec4 ssStart_4 = proj * vec4(csStart, 1), ssEnd_4 = proj * vec4(csEnd, 1);
	vec2 ssStart = ssStart_4.xy / ssStart_4.w, ssEnd = ssEnd_4.xy / ssEnd_4.w;
//	vec2 pxStart = ((ssStart + 1) * .5) * screenW, pxEnd = (1.0f - (ssEnd + 1) * .5) * screenH;

	if( distanceSquared(ssStart, ssEnd) < EPSILON ) ssEnd += 2.0f * EPSILON;
//	if( distanceSquared(pxStart, pxEnd) < EPSILON ) pxEnd += 2.0f * EPSILON;
	//----------------------//
	//----------------------//

	
	//It's faster to precompute the inverse and multiply by it
	// then divide by it normally __IF__ you divide more than once
	const float invMaxSteps = 1.0f / MAX_STEPS;
	

	//-------------------//
	//---"Derivatives"---//
	//-------------------//
	// cs = camera-space
	// ss = screen-space
	// px = pixel-space
	vec3 csDelta = (csEnd - csStart) * invMaxSteps;
	vec2 ssDelta = (ssEnd - ssStart) * invMaxSteps;
//	vec2 pxDelta = (pxEnd - pxStart) * invMaxSteps;
	//-------------------//
	//-------------------//


	
	const vec2 pixelToUV = 1.0f / vec2(screenW, screenH);

	csStart += csDelta * jitter;
	ssStart += ssDelta * jitter;
//	pxStart += pxDelta * jitter;
	vec3 csPos = csStart;
	vec2 ssPos = ssStart;
//	vec2 pxPos = pxStart;

	float rayMaxZ = csRay.z;
	bool found = false;

	for(uint iStep = 0; iStep < MAX_STEPS; ++iStep)
	{
		hitPixel = ssPos;
		hitPixel = (hitPixel + 1.0f) * .5f; //to uv
		hitPixel.y = 1.0f - hitPixel.y; //y has to be flipped

//		hitPixel = pxPos * pixelToUV;


		float rayMinZ = min(csPos.z, rayMaxZ);
		rayMaxZ = max(csPos.z, rayMaxZ);

//		float sceneMaxZ = texelFetch(texNormalDepth, ivec2(hitPixel), 0).w;
		float sceneMaxZ = texture(texNormalDepth, hitPixel).w;
		if( isHyperbolicZBuffer ) sceneMaxZ = linearizeZ(sceneMaxZ, NEAR_Z, FAR_Z);
		float sceneMinZ = sceneMaxZ - Z_THICKNESS;

		if( ((rayMaxZ >= sceneMinZ) && (rayMinZ <= sceneMaxZ)) )
		{
			found = true;
			break; //found our pixel
		}

		ssPos = ssStart + ssDelta * iStep * stepScale;
		csPos = csStart + csDelta * iStep * stepScale;
//		pxPos = pxStart + pxDelta * iStep * stepScale;
	}

	return( found );
}





// Shlick's approximation of the Fresnel factor.
vec3 fresnelSchlick(vec3 F0, float cosTheta)
{
	return F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0);
}
const vec3 Fdielectric = vec3(0.04);



vec2 poissonDisk[5] = vec2[](
	vec2(0),
	vec2(-0.94201624, -0.39906216),
	vec2(0.94558609, -0.76890725),
	vec2(-0.094184101, -0.92938870),
	vec2(0.34495938, 0.29387760)
);

void main()
{
	vec3 originalColor = texture(texBackbuffer, inTex).rgb;

	vec4 albedoMetal = texture(texAlbedoMetal, inTex);
	vec3 albedo = albedoMetal.xyz;
	float metalness = albedoMetal.w;

	vec4 emissiveRoughness = texture(texEmissiveRoughness, inTex);
	vec3 emissive = emissiveRoughness.rgb;
	float roughness = emissiveRoughness.a;

	vec4 normalDepth = texture(texNormalDepth, inTex);
	vec3 normal = normalDepth.xyz;
	float depth = normalDepth.w;
	
	vec4 posH = vec4((inTex.xy - .5) * 2, depth, 1);
	vec4 posW_4 = invVP * posH;
	vec3 posW = posW_4.xyz / posW_4.w; //divide by W = Depth transformation

	vec3 toEye = normalize(eyePos - posW);

	// w = world-space
	// cs = camera-space
	vec3 wRay = normalize(reflect(toEye, normal));
	vec3 csRay = normalize((view * vec4(wRay, 0)).xyz);
	

	
	vec4 hitColor = vec4(0);
	vec2 p = vec2(0);
	float stepScale = 1.0, jitter = 0;
	vec4 csStart_4 = view * vec4(posW, 1);
	vec3 csStart = csStart_4.xyz / csStart_4.w;
	bool reflectionFound = false;


//	if( reflectionFound = raymarch(csStart, csRay, stepScale, jitter, p) )
//		hitColor = texture(texBackbuffer, p);
	reflectionFound = raymarch(csStart, csRay, stepScale, jitter, p);


//	float poissonSclae = 1.0f / max(screenH, screenW);
//
//	uint nFound = 0;
//	for(uint i = 0; i < 1; ++i)
//	for(uint i = 0; i < 5; ++i)
//	{
//		jitter = poissonDisk[i].x * poissonSclae;
//		stepScale = 1.0f + poissonDisk[i].y * poissonSclae;
//
//		if( raymarch(csStart, csRay, stepScale, jitter, p) )
//		{
//			hitColor += texture(texBackbuffer, p);
//			++nFound;
//			reflectionFound = true;
//		}
//	}
//	hitColor /= nFound;
	
	
	
	//if no reflection was found
	// sample our skybox

	vec3 modifiedWRay = normalize(wRay + (invTransView * vec4(0, 0, -1, 0)).xyz);
//	vec3 color = reflectionFound ? hitColor.rgb : vec3(0);
//	vec3 reflectColor = reflectionFound ? hitColor.rgb : texture(texSkybox, csRay).bgr;


	vec3 color = originalColor;

	if( !PASSTHROUGH )
	{
		// Outgoing light direction (vector from world-space fragment position to the "eye").
		vec3 Lo = normalize(eyePos - posW);

		// Angle between surface normal and outgoing light direction.
		float cosLo = max(0.0, dot(normal, Lo));
		
		// Specular reflection vector.
		vec3 Lr = 2.0 * cosLo * normal - Lo;

		// Fresnel reflectance at normal incidence (for metals use albedo color).
		vec3 F0 = mix(Fdielectric, albedo, metalness);

		//IBL lighting from:
		// https://github.com/Nadrin/PBR/blob/cd61a5d59baa15413c7b0aff4a7da5ed9cc57f61/data/shaders/hlsl/pbr.hlsl#L140
		{
			// Calculate Fresnel term for ambient lighting.
			// Since we use pre-filtered cubemap(s) and irradiance is coming from many directions
			// use cosLo instead of angle with light's half-vector (cosLh above).
			// See: https://seblagarde.wordpress.com/2011/08/17/hello-world/
			vec3 F = fresnelSchlick(F0, cosLo);

			// Get diffuse contribution factor (as with direct lighting).
			vec3 kd = mix(1 - F, vec3(0), metalness);
//			vec3 kd = mix(vec3(0), 1 - F, metalness);
//			float ks = max(cosLo, 0.0) * pow(1.0f - roughness, 1000);
//			float spec = (1.0f - roughness);
			float spec = (1 - roughness);
//			float ks = spec;
//			float ks = spec * spec;
//			float ks = spec * spec * spec;
//			float ks = max(cosLo, 0.0) * spec * spec * spec;
			float ks = max(cosLo, 0.0) * spec;


			// Sample pre-filtered specular reflection environment at correct mipmap level.
			vec3 specularIrradiance;
			if( reflectionFound )
				specularIrradiance = textureLod(texBackbuffer, p, roughness * textureQueryLevels(texBackbuffer)).rgb;
//				specularIrradiance = texture(texBackbuffer, p).rgb;
			else
				specularIrradiance = textureLod(texSkybox, Lr, roughness * textureQueryLevels(texSkybox)).bgr;
//			vec3 specularIrradiance = specularTexture.SampleLevel(defaultSampler, Lr, roughness * specularTextureLevels).rgb;


			// Irradiance map contains exitant radiance assuming Lambertian BRDF, no need to scale by 1/PI here either.
//			vec3 diffuseIBL = albedo * specularIrradiance;
//			vec3 diffuseIBL = kd * albedo * specularIrradiance;

			// Split-sum approximation factors for Cook-Torrance specular BRDF.
	//		vec2 specularBRDF = specularBRDF_LUT.Sample(spBRDF_Sampler, float2(cosLo, roughness)).rg;

			// Total specular IBL contribution.
//			vec3 specularIBL = (F0 * specularBRDF.x + specularBRDF.y) * specularIrradiance;

			// Total ambient lighting contribution.
//			vec3 specular = specularIrradiance;
//			vec3 specular = specularIrradiance * ks;
//			color += diffuseIBL + specular;
//			color += diffuseIBL * ks;
			color += albedo * specularIrradiance * ks;
//			color += specularIrradiance;
//			color += specular;
		}
	}


	outColor = vec4(color.rgb, 1);
}
