#version 430

#define PI 3.1415926538
#define EPSILON .00001


//######################################################//
//### Standard inputs/uniforms for every pixelshader ###//
//######################################################//
layout(std140, binding = 0) uniform perFrame
{
	mat4x4	vp;
	mat4x4  invVP;

	mat4x4	view;
	mat4x4	invView;
	mat4x4	invTransView;

	mat4x4	proj;
	mat4x4	invProj;

	vec3	eyePos;
	float	gt;
	float	dt;
	uint	screenW;
	uint	screenH;

};

struct pbrMaterial
{
	vec3 albedo;
	float metal;
	vec3 normal;
	float roughness;
};

struct uberLight
{
	uint	type;
	float	falloffC; //Constant falloff; point & spot
	float	falloffLin; //Linear falloff; point & spot
	float	falloffSq; //Square falloff; point & spot
	
	vec3	posW; //point & spot
	float	ambient; //[0,1]-Range, How much light always reaches an object
	vec3	color; //every
	bool	hasShadow; //Does this light cast shadows?
	vec3	dir; //directional & spot
	uint	shadowIndex; //Index into shadowBuffer.shadows[]. Has to stay here for optimal packing

	//spot only
	float innerAngle; //cosing of that angle
	float outerAngle; //cosine of that angle
};
const uint maxLights = 10;
const uint lightType_point = 0;
const uint lightType_spot = 1;
//const uint lightType_directional = 2;

layout(std140, binding = 1) uniform lightsBuffer
{
	uint numLights;

	uberLight lights[maxLights];
};


struct shadowData
{
	mat4x4	vp;	//Light's view-projection. World-space -> light-space
	mat4x4	invVP; //light's inverse view projection matrix. Light-space -> world-space
};

layout(std140, binding = 2) uniform shadowsBuffer
{
	uint numShadows; //safety purposes
	shadowData shadows[maxLights];
};



layout(location = 1) uniform mat4x4 modelToWorld;
//layout(location = 25) uniform mat4x4 modelInvTrans;

//INPUTS from Vertexshader
layout(location = 0) in vec4 iPosW;
layout(location = 1) in vec4 iPosH;
layout(location = 2) in vec3 iNormal;
layout(location = 3) in vec3 iTangent;
layout(location = 4) in vec3 iBitangent;
layout(location = 5) in vec2 iTex;

//INPUT uniforms used incase no textures are set
layout(location = 5) uniform vec3  iAlbedo;
layout(location = 6) uniform float iMetal;
layout(location = 7) uniform float iRoughness;
layout(location = 8) uniform vec3 iEmissive;

layout(location = 9) uniform bool hasAlbedoTexture;
layout(location = 10) uniform bool hasNormalTexture;
layout(location = 11) uniform bool hasMetalTexture;
layout(location = 12) uniform bool hasRoughnessTexture;
layout(location = 13) uniform bool hasEmissiveTexture;
layout(location = 14) uniform float ALPHA = 1;

//INPUT Textures
layout(binding = 0) uniform sampler2D texAlbedo;
layout(binding = 1) uniform sampler2D texNormal;
layout(binding = 2) uniform sampler2D texMetal;
layout(binding = 3) uniform sampler2D texRoughness;
layout(binding = 4) uniform sampler2D texEmissive;

layout(binding = 5) uniform sampler2DArrayShadow shadowMaps;


//OUTPUTS
//OIT Stuff are also the output
struct perPixelData
{
	vec4	color;
	float	depth;
	uint	prev;
};

layout(binding = 0, r32ui)		uniform uimage2D headPointers;
layout(binding = 0, offset = 0)	uniform atomic_uint nextNodeCounter;
layout(binding = 0, std430)		buffer linkedPixels { perPixelData pixels[]; };
//######################################################//
//### Standard inputs/uniforms for every pixelshader ###//
//######################################################//


//PBR calculations done as seen here:
// https://github.com/Nadrin/PBR/blob/master/data/shaders/hlsl/pbr.hlsl
// 

// GGX/Towbridge-Reitz normal distribution function.
// Uses Disney's reparametrization of alpha = roughness^2.
float ndfGGX(float cosLh, float roughness)
{
	float alpha   = roughness * roughness;
	float alphaSq = alpha * alpha;

	float denom = (cosLh * cosLh) * (alphaSq - 1.0) + 1.0;
	return alphaSq / (PI * denom * denom);
}

// Single term for separable Schlick-GGX below.
float gaSchlickG1(float cosTheta, float k)
{
	return cosTheta / (cosTheta * (1.0 - k) + k);
}

// Schlick-GGX approximation of geometric attenuation function using Smith's method.
float gaSchlickGGX(float cosLi, float cosLo, float roughness)
{
	float r = roughness + 1.0;
	float k = (r * r) / 8.0; // Epic suggests using this roughness remapping for analytic lights.
	return gaSchlickG1(cosLi, k) * gaSchlickG1(cosLo, k);
}

// Shlick's approximation of the Fresnel factor.
vec3 fresnelSchlick(vec3 F0, float cosTheta)
{
	return F0 + (1.0 - F0) * pow(max(0, 1.0 - cosTheta), 5.0);
}



float smootherstep(float edge0, float edge1,  float x)
{
	x = clamp((x - edge0) / (edge1 - edge0), 0.0, 1.0);
	// Evaluate polynomial
	return x * x * x * (x * (x * 6 - 15) + 10);
}

vec3 pbr(pbrMaterial mat, uberLight light, float attenuation, vec3 Lr, vec3 F0, vec3 Li, vec3 Lo, float cosLo)
{
	vec3 directionalLighting = vec3(0);
	{
		vec3 radiance = light.color * clamp(attenuation, 0, 1);

		// Half-vector between Li and Lo.
		vec3 Lh = normalize(Li + Lo);

		// Calculate angles between surface normal and various light vectors.
		float cosLi = max(0.0, dot(mat.normal, Li));
		float cosLh = max(0.0, dot(mat.normal, Lh));

		// Calculate Fresnel term for direct lighting. 
		vec3 F  = fresnelSchlick(F0, max(0, dot(Lh, Lo)));

		// Calculate normal distribution for specular BRDF.
		float D = ndfGGX(cosLh, mat.roughness);

		// Calculate geometric attenuation for specular BRDF.
		float G = gaSchlickGGX(cosLi, cosLo, mat.roughness);

		// Diffuse scattering happens due to light being refracted multiple times by a dielectric medium.
		// Metals on the other hand either reflect or absorb energy, so diffuse contribution is always zero.
		// To be energy conserving we must scale diffuse BRDF contribution based on Fresnel factor & metalness.
		vec3 kd = mix(vec3(1) - F, vec3(0), mat.metal);
		
		// Lambert diffuse BRDF.
		// We don't scale by 1/PI for lighting & material units to be more convenient.
		// See: https://seblagarde.wordpress.com/2012/01/08/pi-or-not-to-pi-in-game-lighting-equation/
		vec3 diffuseBRDF = kd * mat.albedo;

		// Cook-Torrance specular microfacet BRDF.
		vec3 specularBRDF = (F * D * G) / max(EPSILON, 4.0 * cosLi * cosLo) * (mat.roughness);
		directionalLighting = (diffuseBRDF + specularBRDF) * radiance * cosLi;
	}

	// Arbitrary Ambient lighting loosely based on:
	// https://github.com/Nadrin/PBR/blob/cd61a5d59baa15413c7b0aff4a7da5ed9cc57f61/data/shaders/hlsl/pbr.hlsl#L177
	vec3 ambient = vec3(0);
	ambient = mat.albedo * light.ambient;
	{
		// Sample diffuse irradiance at normal direction.
		float irradiance = max(0, dot(Lo, Li));

		// Calculate Fresnel term for ambient lighting.
		// Since we use pre-filtered cubemap(s) and irradiance is coming from many directions
		// use cosLo instead of angle with light's half-vector (cosLh above).
		// See: https://seblagarde.wordpress.com/2011/08/17/hello-world/
		vec3 F = fresnelSchlick(F0, cosLo);

		// Get diffuse contribution factor (as with direct lighting).
		vec3 kd = mix(vec3(1) - F, vec3(0), mat.metal);

		// Irradiance map contains exitant radiance assuming Lambertian BRDF, no need to scale by 1/PI here either.
		vec3 diffuseIBL = kd * mat.albedo * irradiance;

		// Sample pre-filtered specular reflection environment at correct mipmap level.
		float specularIrradiance = 1;

		// Total specular IBL contribution.
		vec3 specularIBL = (F0 * (1 - mat.roughness)) * specularIrradiance;

		// Total ambient lighting contribution.
		ambient = diffuseIBL + specularIBL;
	}

	vec3 color = ambient + directionalLighting;

	return( color );
}

vec3 calcPointlight(const pbrMaterial mat, vec3 posW, uberLight light, vec3 Lr, vec3 F0, vec3 Lo, float cosLo)
{
	const float dst = distance(posW, light.posW);
	const float attenuation = 1 / ((dst * (light.falloffLin + (light.falloffSq * light.falloffSq))) + light.falloffC);
	vec3 Li = normalize(light.posW - posW); //vector from "posW" to light
	return( pbr(mat, light, attenuation, Lr, F0, Li, Lo, cosLo) );
}

vec3 calcSpotlight(const pbrMaterial mat, vec3 posW, uberLight light, vec3 Lr, vec3 F0, vec3 Lo, float cosLo)
{
	vec3 lightDir = -normalize(light.dir);
	vec3 Li = normalize(light.posW - posW);
	const float angle = dot(Li, lightDir);
	float attenuation;
	if( angle < light.outerAngle )
		attenuation = 0; //we're outside the spotlight cone
	else
	{
		//inside the cone
		if( angle > light.innerAngle )
			attenuation = 1; //inside the inner cone
		else
			//in between inner and outer -> interpolate a smoothing factor
			attenuation = smootherstep(light.outerAngle, light.innerAngle, angle);
	}
	const float dst = distance(posW, light.posW);
	attenuation *= 1 / 1 / ((dst * (light.falloffLin + (light.falloffSq * light.falloffSq))) + light.falloffC);
	return( pbr(mat, light, attenuation, Lr, F0, Li, Lo, cosLo) );
}

vec3 calcLighting(pbrMaterial mat, vec3 posW, uberLight light, vec3 Lr, vec3 F0, vec3 Lo, float cosLo)
{
	
	switch( light.type )
	{
	case lightType_point:		return( calcPointlight(mat, posW, light, Lr, F0, Lo, cosLo) );
	case lightType_spot:		return( calcSpotlight(mat, posW, light, Lr, F0, Lo, cosLo) );
	default: return( vec3(0) );
	}
}

vec2 poissonDisk[4] = vec2[](
  vec2( -0.94201624, -0.39906216 ),
  vec2( 0.94558609, -0.76890725 ),
  vec2( -0.094184101, -0.92938870 ),
  vec2( 0.34495938, 0.29387760 )
);

float calcShadowFactor(vec3 posW, shadowData shadow, uint shadowIndex)
{
	vec4 posH = shadow.vp * vec4(posW, 1);
	vec3 posShadowSpace = posH.xyz / posH.w; //perspective divide
	vec3 posShadowSpace_texCoord = posShadowSpace * 0.5 + 0.5;
	const vec2 shadowMapTexelSize = 1.f / textureSize(shadowMaps, 0).xy;

	float vis = 0;
	for(int i = 0; i < 4; ++i)
		vis += texture(shadowMaps, vec4(shadowMapTexelSize * .5 + posShadowSpace_texCoord.xy + poissonDisk[i] * shadowMapTexelSize, shadowIndex, posShadowSpace_texCoord.z + EPSILON));
	return( vis / 4 );
}





void main()
{
	vec3 albedo;
	if( hasAlbedoTexture )
		albedo = texture(texAlbedo, iTex).zyx;
	else
		albedo = iAlbedo;

	vec3 normal = normalize(iNormal);
	if( hasNormalTexture )
	{
		vec3 N = (texture(texNormal, iTex).zyx - .5) * 2;

		vec3 tangent = normalize(iTangent);
		vec3 bitangent = normalize(iBitangent);

		//Transform tangent-space vector N to world-space
		// https://gamedev.stackexchange.com/a/57477

		//tangent, bitangent, normal are in world-space
		// -> transforming tangent-space Normal with it gets you a world-space Normal
		mat3 toWorld = mat3(tangent, bitangent, normal);
		normal = normalize(toWorld * N);
	}

	float metal;
	if( hasMetalTexture )
		metal = texture(texMetal, iTex).x;
	else
		metal = iMetal;

	float roughness;
	if( hasRoughnessTexture )
		roughness = texture(texRoughness, iTex).x;
	else
		roughness = iRoughness;

	vec3 emissive;
	if( hasEmissiveTexture )
		emissive = texture(texEmissive, iTex).zyx;
	else
		emissive  = iEmissive;


	float depth = iPosH.z / iPosH.w;


	//Do lighting
	vec3 color = vec3(0);
	if( numLights == 0 )
	{
		color = albedo;
	}
	else
	{
		color = vec3(0);


		// Fresnel reflectance at normal incidence (for metals use albedo color).
		const vec3 Fdielectric = vec3(.04);
		vec3 F0 = mix(Fdielectric, albedo, metal);

		vec3 posW = iPosW.xyz / iPosW.w;
		// Outgoing light direction (vector from world-space fragment position to the "eye").
		vec3 Lo = normalize(eyePos - posW);

		// Angle between surface normal and outgoing light direction.
		float cosLo = max(0, dot(normal, Lo));

		// Specular reflection vector.
		vec3 Lr = 2.0 * cosLo * normal - Lo;

		pbrMaterial mat;
		mat.albedo = albedo;
		mat.metal = metal;
		mat.normal = normal;
		mat.roughness = roughness;

		for(uint i = 0; i < numLights; ++i)
		{
			vec3 lighting = calcLighting(mat, posW, lights[i], Lr, F0, Lo, cosLo);
			float shadow = 1;
			if( lights[i].hasShadow )
				shadow = calcShadowFactor(posW, shadows[lights[i].shadowIndex], lights[i].shadowIndex);
			color += lighting * shadow;
		}

		color += emissive;
	}


	uint index = atomicCounterIncrement(nextNodeCounter); //increase counter & get old value

	//only add pixel to list if there's room for it left
	if( index < pixels.length() )
	{
		uint prev = imageAtomicExchange(headPointers, ivec2(gl_FragCoord.xy), index); //exchange old pixel value with 'index'
		pixels[index].color = vec4(color, ALPHA);
		pixels[index].depth = depth;
		pixels[index].prev = prev;
	}

}
