#version 430

#define PI 3.1415926538
#define EPSILON .00001



//###################################################//
//### Inputs/uniforms for our gbuffer pixelshader ###//
//###################################################//
layout(std140, binding = 0) uniform perFrame
{
	mat4x4	vp;
	mat4x4  invVP;

	mat4x4	view;
	mat4x4	invView;
	mat4x4	invTransView;

	mat4x4	proj;
	mat4x4	invProj;

	vec3	eyePos;
	float	gt;
	float	dt;
	uint	screenW;
	uint	screenH;

};

layout(binding = 0) uniform sampler2D texAlbedoMetal;
layout(binding = 1) uniform sampler2D texNormalDepth;
layout(binding = 2) uniform sampler2D texEmissiveRoughness;
layout(binding = 3) uniform sampler2DArrayShadow shadowMaps;
layout(binding = 4) uniform samplerCube texSkybox;


struct pbrMaterial
{
	vec3 albedo;
	float metal;
	vec3 normal;
	float roughness;
};

struct uberLight
{
	uint	type;
	float	falloffC; //Constant falloff; point & spot
	float	falloffLin; //Linear falloff; point & spot
	float	falloffSq; //Square falloff; point & spot
	
	vec3	posW; //point & spot
	float	ambient; //[0,1]-Range, How much light always reaches an object
	vec3	color; //every
	bool	hasShadow; //Does this light cast shadows?
	vec3	dir; //directional & spot
	uint	shadowIndex; //Index into shadowBuffer.shadows[]. Has to stay here for optimal packing

	//spot only
	float innerAngle; //cosing of that angle
	float outerAngle; //cosine of that angle
};
const uint maxLights = 10;
const uint lightType_point = 0;
const uint lightType_spot = 1;
//const uint lightType_directional = 2;

layout(std140, binding = 1) uniform lightsBuffer
{
	uint numLights;

	uberLight lights[maxLights];
};


struct shadowData
{
	mat4x4	vp;	//Light's view-projection. World-space -> light-space
	mat4x4	invVP; //light's inverse view projection matrix. Light-space -> world-space
};

layout(std140, binding = 2) uniform shadowsBuffer
{
	uint numShadows; //safety purposes
	shadowData shadows[maxLights];
};


//OIT Stuff
struct perPixelData
{
	vec4	color;
	float	depth;
	uint	prev;
};

layout(binding = 0, r32ui)		uniform uimage2D headPointers;
layout(binding = 0, offset = 0)	uniform atomic_uint nextNodeCounter;
layout(binding = 0, std430)		buffer linkedPixelsBuffer { perPixelData linkedPixels[]; };



layout(location = 0) in vec2 inTex;
layout(location = 1) in vec3 inViewDirection;

out vec4 outColor;
//###################################################//
//### Inputs/uniforms for our gbuffer pixelshader ###//
//###################################################//


//PBR calculations done as seen here:
// https://github.com/Nadrin/PBR/blob/master/data/shaders/hlsl/pbr.hlsl
// 

// GGX/Towbridge-Reitz normal distribution function.
// Uses Disney's reparametrization of alpha = roughness^2.
float ndfGGX(float cosLh, float roughness)
{
	float alpha   = roughness * roughness;
	float alphaSq = alpha * alpha;

	float denom = (cosLh * cosLh) * (alphaSq - 1.0) + 1.0;
	return alphaSq / (PI * denom * denom);
}

// Single term for separable Schlick-GGX below.
float gaSchlickG1(float cosTheta, float k)
{
	return cosTheta / (cosTheta * (1.0 - k) + k);
}

// Schlick-GGX approximation of geometric attenuation function using Smith's method.
float gaSchlickGGX(float cosLi, float cosLo, float roughness)
{
	float r = roughness + 1.0;
	float k = (r * r) / 8.0; // Epic suggests using this roughness remapping for analytic lights.
	return gaSchlickG1(cosLi, k) * gaSchlickG1(cosLo, k);
}

// Shlick's approximation of the Fresnel factor.
vec3 fresnelSchlick(vec3 F0, float cosTheta)
{
	return F0 + (1.0 - F0) * pow(max(0, 1.0 - cosTheta), 5.0);
}



float smootherstep(float edge0, float edge1,  float x)
{
	x = clamp((x - edge0) / (edge1 - edge0), 0.0, 1.0);
	// Evaluate polynomial
	return x * x * x * (x * (x * 6 - 15) + 10);
}

vec3 pbr(pbrMaterial mat, uberLight light, float attenuation, vec3 Lr, vec3 F0, vec3 Li, vec3 Lo, float cosLo)
{
	vec3 directionalLighting = vec3(0);
	{
		vec3 radiance = light.color * clamp(attenuation, 0, 1);

		// Half-vector between Li and Lo.
		vec3 Lh = normalize(Li + Lo);

		// Calculate angles between surface normal and various light vectors.
		float cosLi = max(0.0, dot(mat.normal, Li));
		float cosLh = max(0.0, dot(mat.normal, Lh));

		// Calculate Fresnel term for direct lighting. 
		vec3 F  = fresnelSchlick(F0, max(0, dot(Lh, Lo)));

		// Calculate normal distribution for specular BRDF.
		float D = ndfGGX(cosLh, mat.roughness);

		// Calculate geometric attenuation for specular BRDF.
		float G = gaSchlickGGX(cosLi, cosLo, mat.roughness);

		// Diffuse scattering happens due to light being refracted multiple times by a dielectric medium.
		// Metals on the other hand either reflect or absorb energy, so diffuse contribution is always zero.
		// To be energy conserving we must scale diffuse BRDF contribution based on Fresnel factor & metalness.
		vec3 kd = mix(vec3(1) - F, vec3(0), mat.metal);
//		vec3 kd = vec3(1) - F *  (1 - mat.metal);
		
		// Lambert diffuse BRDF.
		// We don't scale by 1/PI for lighting & material units to be more convenient.
		// See: https://seblagarde.wordpress.com/2012/01/08/pi-or-not-to-pi-in-game-lighting-equation/
		vec3 diffuseBRDF = kd * mat.albedo;

		// Cook-Torrance specular microfacet BRDF.
		vec3 specularBRDF = (F * D * G) / max(EPSILON, 4.0 * cosLi * cosLo) * (mat.roughness);
		directionalLighting = (diffuseBRDF + specularBRDF) * radiance * cosLi;
	}

	// Arbitrary Ambient lighting loosely based on:
	// https://github.com/Nadrin/PBR/blob/cd61a5d59baa15413c7b0aff4a7da5ed9cc57f61/data/shaders/hlsl/pbr.hlsl#L177
	vec3 ambient = vec3(0);
//	ambient = mat.albedo * light.ambient;
//	{
//		// Sample diffuse irradiance at normal direction.
//		float irradiance = max(0, dot(Lo, Li));
//
//		// Calculate Fresnel term for ambient lighting.
//		// Since we use pre-filtered cubemap(s) and irradiance is coming from many directions
//		// use cosLo instead of angle with light's half-vector (cosLh above).
//		// See: https://seblagarde.wordpress.com/2011/08/17/hello-world/
//		vec3 F = fresnelSchlick(F0, cosLo);
//
//		// Get diffuse contribution factor (as with direct lighting).
//		vec3 kd = mix(vec3(1) - F, vec3(0), mat.metal);
//
//		// Irradiance map contains exitant radiance assuming Lambertian BRDF, no need to scale by 1/PI here either.
//		vec3 diffuseIBL = kd * mat.albedo * irradiance;
//
//		// Sample pre-filtered specular reflection environment at correct mipmap level.
//		float specularIrradiance = 1;
//
//		// Split-sum approximation factors for Cook-Torrance specular BRDF.
////		vec3 specularBRDF = specularBRDF_LUT.Sample(spBRDF_Sampler, float2(cosLo, roughness)).rg;
//
//		// Total specular IBL contribution.
//		vec3 specularIBL = (F0 * (1 - mat.roughness)) * specularIrradiance;
//
//		// Total ambient lighting contribution.
//		ambient = diffuseIBL + specularIBL;
//	}

	vec3 color = ambient + directionalLighting;

	return( color );
}

vec3 calcPointlight(const pbrMaterial mat, vec3 posW, uberLight light, vec3 Lr, vec3 F0, vec3 Lo, float cosLo)
{
	const float dst = distance(posW, light.posW);
	const float attenuation = 1 / ((dst * (light.falloffLin + (light.falloffSq * light.falloffSq))) + light.falloffC);
	vec3 Li = normalize(light.posW - posW); //vector from "posW" to light
	return( pbr(mat, light, attenuation, Lr, F0, Li, Lo, cosLo) );
}

vec3 calcSpotlight(const pbrMaterial mat, vec3 posW, uberLight light, vec3 Lr, vec3 F0, vec3 Lo, float cosLo)
{
	vec3 lightDir = -normalize(light.dir);
	vec3 Li = normalize(light.posW - posW);
	const float angle = dot(Li, lightDir);
	float attenuation;
	if( angle < light.outerAngle )
		attenuation = 0; //we're outside the spotlight cone
	else
	{
		//inside the cone
		if( angle > light.innerAngle )
			attenuation = 1; //inside the inner cone
		else
			//in between inner and outer -> interpolate a smoothing factor
			attenuation = smootherstep(light.outerAngle, light.innerAngle, angle);
	}
	const float dst = distance(posW, light.posW);
	attenuation *= 1 / 1 / ((dst * (light.falloffLin + (light.falloffSq * light.falloffSq))) + light.falloffC);
	return( pbr(mat, light, attenuation, Lr, F0, Li, Lo, cosLo) );
}

vec3 calcLighting(pbrMaterial mat, vec3 posW, uberLight light, vec3 Lr, vec3 F0, vec3 Lo, float cosLo)
{
	
	switch( light.type )
	{
	case lightType_point:		return( calcPointlight(mat, posW, light, Lr, F0, Lo, cosLo) );
	case lightType_spot:		return( calcSpotlight(mat, posW, light, Lr, F0, Lo, cosLo) );
	default: return( vec3(0) );
	}
}

vec2 poissonDisk[4] = vec2[](
  vec2( -0.94201624, -0.39906216 ),
  vec2( 0.94558609, -0.76890725 ),
  vec2( -0.094184101, -0.92938870 ),
  vec2( 0.34495938, 0.29387760 )
);

float calcShadowFactor(vec3 posW, shadowData shadow, uint shadowIndex)
{
	vec4 posH = shadow.vp * vec4(posW, 1);
	vec3 posShadowSpace = posH.xyz / posH.w; //perspective divide
	vec3 posShadowSpace_texCoord = posShadowSpace * 0.5 + 0.5;
	const vec2 shadowMapTexelSize = 1.f / textureSize(shadowMaps, 0).xy;

	float vis = 0;
	for(int i = 0; i < 4; ++i)
		vis += texture(shadowMaps, vec4(shadowMapTexelSize * .5 + posShadowSpace_texCoord.xy + poissonDisk[i] * shadowMapTexelSize, shadowIndex, posShadowSpace_texCoord.z - EPSILON));
	return( vis / 4 );
}


#define MAX_PIXELS_PER_BUCKET 75
vec3 applyOIT(const vec3 color, const float depth)
{
	uint index = imageLoad(headPointers, ivec2(gl_FragCoord.xy)).x;
	if( index == uint(-1) ) return( color ); //no OIT pixels

	perPixelData pixels[MAX_PIXELS_PER_BUCKET + 2];
	
	uint numPixels = 1;
	pixels[0].color = texture(texSkybox, inViewDirection).zyxw;
	pixels[0].depth = 1.f;

	if( depth <= 1.0f ) //in case we're sampling a non-written pixel (depth > 1)
	{
		pixels[1].color = vec4(color, 1);
		pixels[1].depth = depth;
		++numPixels;
	}


	for(; numPixels <= MAX_PIXELS_PER_BUCKET && index != uint(-1); ++numPixels)
	{
		pixels[numPixels].color = linkedPixels[index].color;
		pixels[numPixels].depth = linkedPixels[index].depth;
		index = linkedPixels[index].prev;
	}


	//Insert-Sort from:
	// https://github.com/gangliao/Order-Independent-Transparency-GPU/blob/24be55151dac67d3cc82337c22561a337ed38ee2/source%20code/src/shader/oit.fs#L100
	for(uint i = 1; i < numPixels; ++i)
	{
		perPixelData toInsert = pixels[i];
		uint j = i;
		while( j > 0 && toInsert.depth > pixels[j-1].depth ) {
		  pixels[j] = pixels[j-1];
		  j--;
		}
		pixels[j] = toInsert;
	}


	//Apply from back to front
	vec3 oColor = pixels[0].color.rgb;
	for(uint i = 1; i < numPixels; ++i)
	{
		oColor = mix(oColor, pixels[i].color.rgb, pixels[i].color.a);
	}

	return( oColor );
}





void main()
{
	const vec4 albedoMetal = texture(texAlbedoMetal, inTex);
	vec3 albedo = albedoMetal.xyz;
	float metal = albedoMetal.w;

	const vec4 normalDepth = texture(texNormalDepth, inTex);
	vec3 normal = normalDepth.xyz;
	float depth = normalDepth.w;

	const vec4 emissiveRoughness = texture(texEmissiveRoughness, inTex).xyzw;
	vec3 emissive = emissiveRoughness.xyz;
	float roughness = emissiveRoughness.w;

	pbrMaterial mat;
	mat.albedo = albedo;
	mat.metal = metal;
	mat.normal = normal;
	mat.roughness = roughness;

	vec3 color;
	if( numLights == 0 )
	{
		color = albedo;
	}
	else
	{
		color = vec3(0);

		//Since we're rendering a fullscreen-tri
		// "inTex" is the screen coordinate in [0,1]-Range
		// Transform to [-1, 1] = Homogenous Coordinates
		// "Depth" is in homogneous space already [-1,1]
		vec4 posH = vec4((inTex.xy - .5) * 2, depth, 1);
		vec4 posW_4 = invVP * posH;
		vec3 posW = posW_4.xyz / posW_4.w; //divide by W = Depth transformation

		// Fresnel reflectance at normal incidence (for metals use albedo color).
		const vec3 Fdielectric = vec3(.04);
		vec3 F0 = mix(Fdielectric, albedo, metal);

		// Outgoing light direction (vector from world-space fragment position to the "eye").
		vec3 Lo = normalize(eyePos - posW);

		// Angle between surface normal and outgoing light direction.
		float cosLo = max(0, dot(normal, Lo));

		// Specular reflection vector.
		vec3 Lr = 2.0 * cosLo * normal - Lo;

		for(uint i = 0; i < numLights; ++i)
		{
			vec3 lighting = calcLighting(mat, posW, lights[i], Lr, F0, Lo, cosLo);
			float shadow = 1;
			if( lights[i].hasShadow )
				shadow = calcShadowFactor(posW, shadows[lights[i].shadowIndex], lights[i].shadowIndex);
			color += lighting * shadow;
		}

	}

	//make 'emissive' pop out;
	color += emissive * 1.5;
//	color = albedo;
	color = applyOIT(color, depth);

	outColor = vec4(color, 1);
}
